//
// Copyright 2007 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "CoordinateTransform.H"

CoordinateTransform::CoordinateTransform() {
	reset();
}

void
CoordinateTransform::reset() {
	shift_x = shift_y = 0.0;
	scale_x = scale_y = 1.0;
}

void
CoordinateTransform::copy(CoordinateTransform *c) {
	shift_x = c->shift_x;
	shift_y = c->shift_y;
	scale_x = c->scale_x;
	scale_y = c->scale_y;
}

void
CoordinateTransform::combine(CoordinateTransform *c) {
	scale_x = scale_x * c->scale_x;
	shift_x = shift_x * c->scale_x + c->shift_x;
	scale_y = scale_y * c->scale_y;
	shift_y = shift_y * c->scale_y + c->shift_y;
}

double
CoordinateTransform::transform_x(double _v, dir_t d) {
	if (d == TO_LOCAL)
		return _v * scale_x + shift_x;
	else
		return (_v - shift_x) / scale_x;
}

double
CoordinateTransform::transform_y(double _v, dir_t d) {
	if (d == TO_LOCAL)
		return _v * scale_y + shift_y;
	else
		return (_v - shift_y) / scale_y;
}

double
CoordinateTransform::transform_scale_x(double _v, dir_t d) {
	if (d == TO_LOCAL)
		return _v * scale_x;
	else
		return _v / scale_x;
}

double
CoordinateTransform::transform_scale_y(double _v, dir_t d) {
	if (d == TO_LOCAL)
		return _v * scale_y;
	else
		return _v / scale_y;
}

static int
check_float(const char *str) {
	if (*str == '-')
		str++;
	while (isdigit(*str))
		str++;
	if (*str == '.' && isdigit(*(str+1)))
		return 1;
	else
		return 0;
}

static const char *
end_of_number(const char *str) {
	if (*str == '-') str++;
	while (isdigit(*str)) str++;
	if (*str == '.' && isdigit(*(str+1))) {
		str++;
		while (isdigit(*str)) str++;
	}

	return str;
}

int
CoordinateTransform::convert(CoordinateTransform *loc_trans,
	ImageInfo *img_info,
	const char *str, char *buf, int buflen, dir_t d, int strip) {
	loc_trans->reset();
	return convert_rec(loc_trans, img_info, str, buf, buflen, d, strip);
}

int
CoordinateTransform::convert_rec(CoordinateTransform *loc_trans, 
	ImageInfo *img_info,
	const char *str, char *buf, int buflen, dir_t d, int strip) {

	static const char *prefix[] = {"_x", "_ox", "_X", "_y", "_oy", "_Y",
		"_h", "_H", "_v", "_V", "_S", "_fx", "_fy"};
	int type = -1, is_float;
	double value;

	if (strlen(str) < 3) {
		strncpy(buf, str, buflen);
		return 0;
	}

	for (unsigned int i=0; i < sizeof(prefix)/sizeof(prefix[0]); i++) {
		if (strncmp(str, prefix[i], strlen(prefix[i])) == 0 &&
			(isdigit(*(str + strlen(prefix[i]))) ||
			(*(str + strlen(prefix[i])) == '-' && strlen(str) > 3 &&
			isdigit(*(str + strlen(prefix[i]) + 1))))) {
	
			type = i;
			str += strlen(prefix[i]);
			break;
		}
	}

	if (type == -1) {
		*buf = *str++;
		buflen -= strlen(buf);
		buf += strlen(buf);

		return convert_rec(loc_trans, img_info, str, buf, buflen, d, strip);
	}

	if (!strip) {
		strncat(buf, prefix[type], buflen);
		buflen -= strlen(buf);
		buf += strlen(buf);
	}

	value = atof(str);
	is_float = check_float(str);

	switch (type) {
		case 0:
			value = transform_x(value, d);
			break;
		case 1:
			loc_trans->shift_x = value;
			break;
		case 2:
			value = transform_x(value, d);
			loc_trans->shift_x = -value;
			break;
		case 3:
			value = transform_y(value, d);
			break;
		case 4:
			loc_trans->shift_y = value;
			break;
		case 5:
			value = transform_y(value, d);
			loc_trans->shift_y = -value;
			break;
		case 6:
			value = transform_scale_x(value, d);
			break;
		case 7:
			loc_trans->scale_x = value;
			break;
		case 8:
			value = transform_scale_y(value, d);
			break;
		case 9:
			loc_trans->scale_y = value;
			break;
		case 10:
			if (!is_float) {
				// assume percent value
				loc_trans->scale_x = value / 100.0;
				loc_trans->scale_y = value / 100.0;
			} else {
				loc_trans->scale_x = value;
				loc_trans->scale_y = value;
			}
			break;
		case 11:
			if (img_info && img_info->format != ImageInfo::UNKNOWN) {
				double sc = value / (double) img_info->width;
				if (sc < loc_trans->scale_x) {
					loc_trans->scale_x = sc;
					loc_trans->scale_y = sc;
				}
			} else
				fprintf(stderr, "Warning: No size information available\n");
			break;
		case 12:
			if (img_info && img_info->format != ImageInfo::UNKNOWN) {
				double sc = value / (double) img_info->height;
				if (sc < loc_trans->scale_y) {
					loc_trans->scale_x = sc;
					loc_trans->scale_y = sc;
				}
			} else
				fprintf(stderr, "Warning: No size information available\n");
			break;
		default:
			break;
	}

	if (is_float)
		snprintf(buf, buflen, "%f", value);
	else
		snprintf(buf, buflen, "%d", (int) rint(value));

	buflen -= strlen(buf);
	buf += strlen(buf);

	str = end_of_number(str);

	return convert_rec(loc_trans, img_info, str, buf, buflen, d, strip);
}
