//
// Copyright 2007-2010 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <FL/Fl.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/Fl_Pack.H>
#include <FL/fl_draw.H>
#include "FlimpWidget.H"
#include "FlimpText.H"
#include "Rectangle.H"

class Scrollbar : public Fl_Scrollbar {
	public:
		Scrollbar(int _x, int _y, int _w, int _h) :
			Fl_Scrollbar(_x, _y, _w, _h) {};

		int handle (int event) {
			if ((event == FL_KEYBOARD || event == FL_SHORTCUT) &&
				(Fl::event_key() == FL_Left ||
				 Fl::event_key() == FL_Right ||
				 Fl::event_key() == FL_Up ||
				 Fl::event_key() == FL_Down)) {
				return 0;
			} else {
				return Fl_Scrollbar::handle(event);
			}
		}
};

double FlimpWidget::scroll_center_x;
double FlimpWidget::scroll_center_y;
double FlimpWidget::scroll_center_pos_x;
double FlimpWidget::scroll_center_pos_y;

void
FlimpWidget::up() {
	iq->up();
	iq->process();
}

void
FlimpWidget::down() {
	iq->down();
	iq->process();
}

FlimpWidget::FlimpWidget(int _x, int _y, int _w, int _h, const char *n) :
	Fl_Group(_x, _y, _w, _h) {
	Scrollbar *sb_h, *sb_v;
	int top_border = 1, control_h = 25;

	scroll = new Fl_Huge_Scroll(_x, _y + top_border,
		_w - 10, _h - control_h - 10 - top_border);
	scroll->end();
	scroll->color(FL_WHITE);
	img = new SelectImage(_x, _y + top_border,
		_w - 10, _h - control_h - 10 - top_border);
	img->set_type(SelectImage::NONE);
	scroll->add(img, 0, 0);
	sb_h = new Scrollbar(_x, _y + _h - control_h - 10, _w - 10, 10);
	sb_h->type(FL_HORIZONTAL);
	scroll->set_horizontal_scrollbar(sb_h);
	sb_v = new Scrollbar(_x + _w - 10, _y + top_border,
		10, _h - control_h - 10 - top_border);
	sb_v->type(FL_VERTICAL);
	scroll->set_vertical_scrollbar(sb_v);
	scroll->register_user_scroll_callback(user_scroll_callback, this);
	resizable(scroll);
	
	ctrl = new Fl_Pack(_x + 1, _y + _h - control_h + 2, _w, control_h - 4);
	ctrl->type(Fl_Pack::HORIZONTAL);
	ctrl->spacing(2);

	iq = new ImageQueue();
	iq->register_proc_callback(FlimpWidget::process_callback, this);
	filename = NULL;
	tabs = NULL;

	copy_label(n?n:"Generic");
}

FlimpWidget::~FlimpWidget() {
	if (iq)
		delete iq;
	if (filename)
		free(filename);
}

void
FlimpWidget::adjust_scroll() {
	double X = iq->global_trans->transform_x(scroll_center_x,
		CoordinateTransform::TO_LOCAL);
	double Y = iq->global_trans->transform_y(scroll_center_y,
		CoordinateTransform::TO_LOCAL);

	scroll->position((int) rint(X - scroll_center_pos_x),
		(int) rint(Y - scroll_center_pos_y));
}

int
FlimpWidget::handle(int event) {
	if (event == FL_SHOW) {
		iq->process();
		adjust_scroll();
	}

	return Fl_Group::handle(event);
}

int
FlimpWidget::load(const char *p) {
	int ret;

	if (p == NULL)
		return 1;
	
	if (!filename) {
		filename = strdup(p);
		iq->set_file(filename);
	} else if (strcmp(filename, p)) {
		free(filename);
		filename = strdup(p);
		iq->set_file(filename);
	} 

 	ret = img->load(filename);
	scroll->redraw();
	return ret;	
}

int
FlimpWidget::reload() {
	if (filename != NULL)
		return load(filename);
	else
		return 1;
}

void
FlimpWidget::set_local_command(const char *c) {
	if (tabs)
		tabs->value(this);
	iq->set_local_command(c);
}

void
FlimpWidget::add_control(Fl_Widget *w) {
	ctrl->insert(*w, 0);
	if (ctrl->children() == 1)
		ctrl->resizable(w);
}

void
FlimpWidget::store_visible_center() {
	// The user has actively modified the scroll position or the image size.
	// Store the center of the visible image part in original coordinates
	// together with its current position on the screen.

	Rectangle ri(0, 0, img->w(), img->h());
	Rectangle rs(scroll->xposition(), scroll->yposition(),
		scroll->w(), scroll->h());

	ri.intersect(&rs);

	scroll_center_x = (double) ri.x() + ri.w() / 2.0;
	scroll_center_y = (double) ri.y() + ri.h() / 2.0;

	scroll_center_pos_x = scroll_center_x - scroll->xposition();
	scroll_center_pos_y = scroll_center_y - scroll->yposition();

	scroll_center_x = iq->global_trans->transform_x(scroll_center_x,
        CoordinateTransform::TO_ORIGINAL);
	scroll_center_y = iq->global_trans->transform_y(scroll_center_y,
		CoordinateTransform::TO_ORIGINAL);
}

void
FlimpWidget::process_callback(void *d) {
	FlimpWidget *fl = (FlimpWidget *)d;

	fl->set_command(fl->iq->get_local_command());
	fl->reload();
	fl->adjust_scroll();
	fl->store_visible_center();
}


void
FlimpWidget::user_scroll_callback(void *d) {
	FlimpWidget *fl = (FlimpWidget*) d;

	fl->store_visible_center();
}
