//
// Copyright 2007 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "ImageQueue.H"

ImageQueue::ImageQueue() {
	static int id_gen;

	filename = NULL;
	prev = next = NULL;
	id = id_gen++;
	memset(global_command, 0, sizeof(global_command));	
	local_trans  = new CoordinateTransform();
	global_trans = new CoordinateTransform();
	img_info = new ImageInfo();
	proc_callback = NULL;
	proc_callback_data = NULL;
	processing_needed = 0;
	generation = 0;
	processed_generation = -1;
	processed_id = -1;
}

ImageQueue::~ImageQueue() {
	if (prev && filename && strncmp(filename, "/tmp/flimp", 10) == 0)
		unlink(filename);
	remove();
	delete local_trans;
	delete global_trans;
	delete img_info;
	if (filename)
		free(filename);
}

void
ImageQueue::set_file(const char *f) {
	if (!f)
		return;
	if (filename)
		free(filename);
	filename = strdup(f);
	processing_needed++;
}

int
ImageQueue::convert(const char *str, char *buf, int buflen,
	CoordinateTransform::dir_t d, int strip) {
	CoordinateTransform dummy_gt, *gt;

	if (prev)
		gt = prev->global_trans;
	else
		gt = &dummy_gt;

	memset(buf, 0, buflen);
	return gt->convert(local_trans, prev?prev->img_info:NULL,
		str, buf, buflen, d, strip);
}

void
ImageQueue::set_local_command(const char *c) {
	if (c == NULL)
		return;

	convert(c, global_command, sizeof(global_command),
		CoordinateTransform::TO_ORIGINAL, 0);
	
	processing_needed++;
	process();
}

void
ImageQueue::set_global_command(const char *c) {
	if (c == NULL)
		return;
	
	strncpy(global_command, c, sizeof(global_command));
	processing_needed++;
}

const char*
ImageQueue::get_local_command() {
	convert(global_command, local_command, sizeof(local_command),
		CoordinateTransform::TO_LOCAL, 0);

	return local_command;
}

int
ImageQueue::process() {
	int ret;
	char *f, buf[1024];

	if (prev != NULL) {
		ret = prev->process();

		if ((processing_needed ||
				ret != processed_generation ||
				processed_id != prev->id) &&
			global_command[0] != '\0') {
			int st;

			convert(global_command, local_command, sizeof(local_command),
				CoordinateTransform::TO_LOCAL, 1);

			update_global_trans();

			unlink(get_file());

			if (strstr(local_command, "_$")) {
				char buf1[1024];
				if (!prev) {
					fprintf(stderr, "_$ not supported in first command\n");
					return 1;
				}

				strncpy(buf1, local_command, sizeof(buf1));
				f = strstr(buf1, "_$");
				*f = '\0';

				snprintf(buf, sizeof(buf), "(%s'%s'%s) > '%s'",
					buf1, prev->get_file(), f+2, get_file());

			} else
				snprintf(buf, sizeof(buf), "(%s) < '%s' > '%s'",
					local_command, prev->get_file(), get_file());

			fprintf(stderr, "%s", local_command);
			st = system(buf);
			st = WEXITSTATUS(st);
			fprintf(stderr, " returned %d\n", st);
			generation++;
			processed_generation = ret;
			processed_id = prev->id;

			img_info->load(filename);

			if (proc_callback != NULL)
				proc_callback(proc_callback_data);
		}
	} else if (processing_needed) {
		generation++;
		if (proc_callback != NULL)
			proc_callback(proc_callback_data);
	}

	processing_needed = 0;

	return generation;
}

void
ImageQueue::update_global_trans() {
	if (prev)
		global_trans->copy(prev->global_trans);
	else
		global_trans->reset();

	global_trans->combine(local_trans);
}

void
ImageQueue::append(ImageQueue *q) {
	if (next == NULL) {
		next = q;
		q->prev = this;
		q->next = NULL;
	} else
		next->append(q);

	q->update_global_trans();
}

void
ImageQueue::insert(ImageQueue *q) {
	ImageQueue *t_next = next;

	next = q;
	q->prev = this;
	q->next = t_next;
	if (t_next)
		t_next->prev = q;
	
	q->update_global_trans();
}

void
ImageQueue::remove() {
	if (prev)
		prev->next = next;
	if (next)
		next->prev = prev;

	prev = next = NULL;
}

void
ImageQueue::up() {
	ImageQueue *t_next = next;
	if (next) {
		remove();
		t_next->insert(this);
	}
}

void
ImageQueue::down() {
	ImageQueue *t_prev = prev->prev;
	if (t_prev) {
		remove();
		t_prev->insert(this);
	}
}

void
ImageQueue::register_proc_callback(Callback *cb, void *d) {
	proc_callback = cb;
	proc_callback_data = d;
}
