//
// Copyright 2007-2010 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <libgen.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Tabs.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/Fl_File_Chooser.H>
#include <FL/Fl_Shared_Image.H>
#include <FL/fl_draw.H>

#include "../config.h"

#include "FlimpWidget.H"
#include "FlimpText.H"
#include "FlimpSelect.H"
#include "ExternalGui.H"

#define WIDTH  600
#define HEIGHT 400 

#ifndef DATADIR
#define DATADIR "/usr/local/share"
#endif
#define DEFAULT_RC_FILE DATADIR "/" PACKAGE_NAME "/flimprc.sample"

static char *default_flimp_file;
static char *default_image_file;
static char *open_image_dir;
static char *save_image_dir;
static char tmpdir[1024];
static Fl_Window *win;
static Fl_Tabs *tabs;
static Fl_Menu_Bar *mb;
static ImageQueue *iq;
static FlimpWidget *flimp;

static void tab_new_cb(Fl_Widget* w, void* d);

struct guidef { char *command, *gui, *gui_data; }; 

static void
usage() {
	fprintf(stderr,
		"usage: flimp [-f <FlimpFile>] <ImageFile>\n"
		" -f <FlimpFile> load commands from file.\n"
		" -t <tab>       jump to tab number <tab> (0 is first, -1 is last).\n"
		" -o <outfile>   set default output file name.\n"
		"additionally flimp supports the following FLTK standard options:\n"
		" -geometry WxH+X+Y\n"
		" -iconic\n"
		" -scheme string\n");
	exit(1);
}

static void
set_file_name(const char *f) {
	char *dot, *img_file = strdup(f);

	if ((dot = strrchr(img_file, '.')))
		*dot = '\0';
	if (default_flimp_file)
		free(default_flimp_file);
	asprintf(&default_flimp_file, "%s.flimp", img_file);
	free(img_file);
	win->copy_label(f);
}

static FlimpWidget *
new_widget(int x, int y, int w, int h,
	const char *name, const char *gui, const char *gui_data) {
	FlimpWidget *fl;

	if (gui && strcmp(gui, "PointSelect") == 0)
		fl = new FlimpSelect(0, 50, win->w(), win->h() - 50,
			SelectImage::POINT, name);
	else if (gui && strcmp(gui, "RectangleSelect") == 0)
		fl = new FlimpSelect(0, 50, win->w(), win->h() - 50,
			SelectImage::RECTANGLE, name);
	else if (gui && strcmp(gui, "ExternalGui") == 0)
		fl = new ExternalGui(0, 50, win->w(), win->h() - 50, name);
	else
		fl = new FlimpText(0, 50, win->w(), win->h() - 50, name);

	fl->end();
	fl->set_tabs(tabs);
	fl->set_gui_data(gui_data);
	return fl;
}

static FlimpWidget*
insert_tab(const char *name, const char *gui, const char *gui_data, FlimpWidget *where) {
	static int count;
	char buf[1024];
	FlimpWidget *fl;

	if (tabs->children() >= 127) {
		fprintf(stderr, "Maximum number of tabs exceeded\n");
		return NULL;
	}

	fl = new_widget(0, 50, win->w(), win->h() - 50, name, gui, gui_data);
	snprintf(buf, sizeof(buf), "%s/flimp_%d", tmpdir, count++);
	tabs->resizable(fl);
	fl->load(buf);

	if (where) {
		tabs->insert(*fl, tabs->find(where) + 1);
		where->iq->insert(fl->iq);
	} else {
		tabs->add(fl);
		iq->append(fl->iq);
	}

	tabs->redraw_label();
	return fl;
}

static int
open_file(const char *file, int menu) {
	char buf[1024];
	FILE *fp = fopen(file, "r");

	if (fp == NULL) {
		if (errno != ENOENT)
			perror("fopen");
		return 1;
	}

	while (fgets(buf, sizeof(buf), fp)) {
		FlimpWidget *fl;
		char *nl = strchr(buf, '\n');
		char *name, *command, *gui = NULL, *gui_data = NULL;

		if (buf[0] == '#')
			continue;
			
		if (nl)
			*nl = '\0';

		name = buf;
		command = strchr(buf, '\t');
		if (command) {
			*command++ = '\0';
		} else {
			fprintf(stderr, "Error parsing %s\n", file);
			fclose(fp);
			return 1;
		}

		gui = strchr(command, '\t');
        if (gui) {
            *gui++ = '\0';
			gui_data = strchr(gui, '\t');
			if (gui_data)
				*gui_data++ = '\0';
		}

		if (menu) {
			struct guidef *g = (struct guidef*) malloc(sizeof(struct guidef));
			g->command = strdup(command);
			g->gui = gui?strdup(gui):NULL;
			g->gui_data = gui_data?strdup(gui_data):NULL;
			mb->add(name, 0, (Fl_Callback*)tab_new_cb, (void*) g);
		} else {
			fl = insert_tab(name, gui, gui_data, NULL);
			if (fl)
				fl->iq->set_global_command(command);
		}
	}

	fclose(fp);
	return 0;
}

static void
save_file(const char *file) {
	FILE *fp = fopen(file, "w");
	int n = tabs->children();

	if (fp == NULL) {
		perror("fopen");
		return;
	}

	for (int i = 1; i < n; i++) {
		FlimpWidget *fl = (FlimpWidget*) tabs->child(i);

		fprintf(fp, "%s\t%s\t%s\t%s\n",
			fl->get_name(), fl->iq->get_global_command(),
			fl->get_gui(), fl->get_gui_data());
	}	

	fclose(fp);
}

static int
confirm_overwrite(const char *f) {
	struct stat sb;
	if (stat(f, &sb) == 0)
		return fl_choice("The file exists.\n", "Cancel", "Overwrite", NULL);
	else
		return 1;
}

static void
quit_cb() {
	if (Fl::event() == FL_SHORTCUT && Fl::event_key() == FL_Escape) 
		return; // ignore Escape

	while (tabs->children() > 0) {
		FlimpWidget *fl = (FlimpWidget *) tabs->child(tabs->children() - 1);
		tabs->remove(*fl);
		delete fl;
	}

	if (rmdir(tmpdir) != 0)
		perror("rmdir");
	exit(0);
}

static void
open_cb() {
	char *file = fl_file_chooser("Open Image?", NULL, open_image_dir);
	if (file != NULL) {
		flimp->load(file);
		set_file_name(file);
		if (open_image_dir)
			free(open_image_dir);
		open_image_dir = strdup(dirname(file));
		FlimpWidget *fl = (FlimpWidget *) tabs->value();
		fl->iq->process();
	}  
}

static void
open_flimp_cb() {
	struct stat sb;
	char *file = fl_file_chooser("Open File?", "*.flimp",
		stat(default_flimp_file, &sb) == 0 ? default_flimp_file : NULL);
	if (file != NULL)
		open_file(file, 0);
}

static void
save_flimp_cb() {
	char *file = fl_file_chooser("Save File?", "*.flimp", default_flimp_file);
	if (file != NULL && confirm_overwrite(file))
		save_file(file);
}

static void
save_image_cb() {
	char *file = fl_file_chooser("Save Image?", NULL,
		default_image_file ? default_image_file : save_image_dir);

	if (file != NULL && confirm_overwrite(file)) {
		FlimpWidget *fl = (FlimpWidget *) tabs->value();
		int r, in_fd, out_fd;
		char buf[1024];

		in_fd = open(fl->iq->get_file(), O_RDONLY);
		if (in_fd < 0) {
			perror("open");
			return;
		}
		out_fd = open(file, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP);
		if (out_fd < 0) {
			perror("open");
			close(in_fd);
			return;
		}

		while ((r = read(in_fd, buf, sizeof(buf))) > 0)
			write(out_fd, buf, r);

		close(in_fd);
		close(out_fd);

		if (save_image_dir)
			free(save_image_dir);
		save_image_dir = strdup(dirname(file));
	}
}

static void
tab_new_cb(Fl_Widget* w, void* d) {
	Fl_Menu_* mw = (Fl_Menu_*)w;
	const Fl_Menu_Item* m = mw->mvalue();
	struct guidef *g = (struct guidef *) d; 
	FlimpWidget *fl;
	FlimpWidget *cur = (FlimpWidget *) tabs->value();

	if (m == NULL || m->label() == NULL)
		return;

	// use strdup() as POSIX basename doesn't have const char* argument
	char *labelbuf = strdup(m->label());
	fl = insert_tab(basename(labelbuf), g?g->gui:NULL, g?g->gui_data:NULL, cur);
	free(labelbuf);

	if  (fl) {
		fl->set_local_command(g?g->command:NULL);
		fl->iq->process();
	}
}

static void
tab_remove_cb() {
	FlimpWidget *fl = (FlimpWidget *) tabs->value();

	if (fl == NULL)
		return;

	int index = tabs->find(fl);
	if (index < 1)
		return;

	tabs->remove(fl);
	tabs->value(tabs->child(index - 1));

	if (tabs->resizable() == fl)
		tabs->resizable(tabs->child(tabs->children() - 1));

	delete fl;
}

static void
tab_up_cb() {
	FlimpWidget *fl = (FlimpWidget *) tabs->value();

	if (fl == NULL)
		return;
	int index = tabs->find(fl);
	if (index <= 0)
		return;

	fl->up();
	tabs->insert(*fl, index + 2);
	win->redraw();
}

static void
tab_down_cb() {
	FlimpWidget *fl = (FlimpWidget *) tabs->value();

	if (fl == NULL)
		return;
	int index = tabs->find(fl);
	if (index <= 1)
		return;

	fl->down();
	tabs->insert(*fl, index - 1);
	win->redraw();
}

static void
about_cb() {
	fl_message("flimp -- a Fast Light Image Processor.\n"
		"Version %s\n\n"
		"(c) Johannes Hofmann 2007 - 2010\n\n",
		VERSION);
}

static void
fill_menubar(Fl_Menu_Bar* mb) {
	char buf[1024];
	
	mb->add("&File/&Open Image", FL_CTRL+'o', (Fl_Callback*)open_cb);
	mb->add("&File/Save &Image", FL_CTRL+'i', (Fl_Callback*)save_image_cb);
	mb->add("&File/Open &Flimp File", FL_CTRL+'f', (Fl_Callback*)open_flimp_cb);
	mb->add("&File/&Save Flimp File", FL_CTRL+'s', (Fl_Callback*)save_flimp_cb);
	mb->add("&File/&Quit", FL_CTRL+'q', (Fl_Callback*)quit_cb);

	mb->add("&Tabs/Generic", FL_CTRL+'n', (Fl_Callback*)tab_new_cb);
	mb->add("&Tabs/&Remove", FL_CTRL+'r', (Fl_Callback*)tab_remove_cb);
	mb->add("&Tabs/&Up", FL_CTRL+'u', (Fl_Callback*)tab_up_cb);
	mb->add("&Tabs/&Down", FL_CTRL+'d', (Fl_Callback*)tab_down_cb);

	snprintf(buf, sizeof(buf), "%s/.flimprc", getenv("HOME"));
	if (open_file(buf, 1) != 0)
		open_file(DEFAULT_RC_FILE, 1);

	mb->add("Help/About", 0, (Fl_Callback*)about_cb);
}

int
main(int argc, char **argv) {
	int i, n, c, tab_num = 0, err = 0, my_argc;
	char **my_argv;
	
	strncpy(tmpdir, "/tmp/flimp.XXXXXX", sizeof(tmpdir));
	if (mkdtemp(tmpdir) == NULL) {
		perror("mkdtemp");
		exit(1);
	}

	fl_register_images();

	win = new Fl_Double_Window(WIDTH, HEIGHT);
	win->callback((Fl_Callback*) quit_cb);
	mb = new Fl_Menu_Bar(0, 0, WIDTH, 30);
	fill_menubar(mb);

	tabs = new Fl_Tabs(0, 30, WIDTH, HEIGHT - 30);
	tabs->selection_color(FL_WHITE);
	tabs->labelcolor(FL_BLACK);
	tabs->end();
	win->end();
	win->resizable(tabs);

	flimp = new FlimpWidget(0, 50, WIDTH, HEIGHT - 50, "Original");
	tabs->add(flimp);
	tabs->resizable(flimp);
	iq = flimp->iq;

	while ((c = getopt(argc, argv, "f:h:g:i:o:s:t:")) != -1) {
		switch (c) {  
			case 'f':
				open_file(optarg, 0);
				break;
			case 'h':
				usage();
				break;
			case 'o':
				default_image_file = optarg;
				break;
			case 't':
				tab_num = atoi(optarg);
				break;
			default:
				i = optind -1;
				n = Fl::arg(argc, argv, i);
				if (n == 0)
					err++;
				else
					optind = i;
		}
	}

	if (err)
		usage();

	my_argc = argc - optind;
	my_argv = argv + optind;

	if (my_argc > 0) {
		set_file_name(my_argv[0]);
		open_image_dir = strdup(dirname(my_argv[0]));
		flimp->load(my_argv[0]);
		if (tab_num > 0)
			tabs->value(tabs->child(tab_num));
		else if (tab_num < 0)
			tabs->value(tabs->child(tabs->children() + tab_num));
	}
	win->end();
	win->show(1, argv);
	signal(SIGPIPE, SIG_IGN);
	return Fl::run();
}
