//
// Copyright 2008 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Counter.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Dial.H>

#define MAX_VAL 1024
#define FORMAT_NUMBER -1

static char cmd[2048];
static int format[2048];
static Fl_Valuator *valuators[MAX_VAL];
static int num_vals;

static void
usage() {
        fprintf(stderr,
                "usage: flimpgui -cdns [<title>] [-r <from>,<to>]\n"
                " -c <title> create counter widget\n"
                " -d <title> create dial widget\n"
                " -n create dummy (invisible) widget\n"
                " -s <title> create slider widget\n"
                " -r <from>,<to> set range for previous widget\n");
        exit(1);
}

static void
gui_cb(Fl_Widget* o, void* d) {
	int n = 0;

	for (int i = 0; format[i]; i++) {
		if (format[i] == FORMAT_NUMBER) {
			while (!valuators[n] && n < num_vals)
				n++;
			printf("%.4g", valuators[n]?valuators[n]->value():0.0);
			n++;
		} else
			printf("%c", format[i]);
	}

	fflush(stdout);
}

static void
stdin_cb(int fd, void *d) {
	int j = 0, n = 0;

 	if (fgets(cmd, sizeof(cmd), stdin) == NULL)
		exit(0);

	for (int i = 0; cmd[i]; i++) {
		char *endptr;
		double v = strtod(&cmd[i], &endptr);

		if (n >= num_vals || isspace(cmd[i]) || cmd[i] == '+' || endptr == &cmd[i]) {
			format[j++] = cmd[i];
		} else {
			if (valuators[n]) {
				format[j++] = FORMAT_NUMBER;
				valuators[n]->value(v);
			} else {
				while(i < endptr - cmd)
					format[j++] = cmd[i++];
			}
			i = endptr - cmd - 1;
			n++;
		}
	}
}

int
main(int argc, char **argv) {
	int c, height = 10;
	Fl_Dial *d;
	Fl_Window window(200, 400, "flimpgui");
	Fl_Pack pack(10, 10, window.w() - 20, window.h() - 30);
	pack.spacing(20);

	while ((c = getopt(argc, argv, "ns:c:d:r:h")) != EOF) {
		switch (c) {  
			case 'n':
				num_vals++;
				break;
			case 'c':
				valuators[num_vals] = new Fl_Counter(0, 0, pack.w(), 25, optarg);
				valuators[num_vals]->callback(gui_cb); 
				height += valuators[num_vals]->h() + pack.spacing();
				num_vals++;
				break;
			case 's':
				valuators[num_vals] = new Fl_Value_Slider(0, 0, pack.w(), 25, optarg);
				valuators[num_vals]->type(FL_HOR_NICE_SLIDER);
				valuators[num_vals]->when(FL_WHEN_RELEASE); 
				valuators[num_vals]->callback(gui_cb); 
				height += valuators[num_vals]->h() + pack.spacing();
				num_vals++;
				break;
			case 'd':
				d = new Fl_Dial(0, 0, pack.w(), pack.w(), optarg);
				d->angles(0, 360);
				valuators[num_vals] = d;
				valuators[num_vals]->type(FL_LINE_DIAL);
				valuators[num_vals]->when(FL_WHEN_RELEASE); 
				valuators[num_vals]->callback(gui_cb); 
				height += valuators[num_vals]->h() + pack.spacing();
				num_vals++;
				break;
			case 'r':
				double from, to;
				from = to = atof(optarg);
				if (strchr(optarg, ','))
					to = atof(strchr(optarg, ',') + 1);
				if (num_vals && valuators[num_vals - 1])
					valuators[num_vals - 1]->range(from, to);
				break;
			default:
				usage();
				break;
		}
		if (num_vals >= MAX_VAL) {
			fprintf(stderr, "Error: Maximum number of widgets (%d) exeeded.\n", MAX_VAL);
			num_vals = MAX_VAL - 1;
		}
	}

	window.resize(window.x(), window.y(), window.w(), height);
	window.show(1, argv);
	Fl::add_fd(0, FL_READ, stdin_cb);

	return Fl::run();
}
